## Replacement gsl::dilog, which is faster, but this way we don't need to
## install gsl
dilog <- function(t)
    vapply(t, function(u) -integrate(function(t)
        log(1-t)/t, 0, u)$value, numeric(1))

#' DGP for simulation design
#' @param n number of observations
#' @param M bound on second derivative of f(X)
#' @param par list with following components:
#' \describe{
#'
#' \item{b1, b2, b3, b4}{Knots for spline. For symmetric designs, \code{p4=NA},
#' and the other parameters range in \eqn{[0, 1]}. For asymmetric designs, the
#' parameters range in \eqn{[-1, 1]}}
#'
#' \item{k}{Tuning parameter to approximate (x)^2_{+} by a smooth function}
#'
#' \item{unif}{\eqn{Unif(-1,1)} or scaled \eqn{Beta(2, 5)} distribution for running
#' variable?}
#'
#' \item{eps}{Integer specifying distribution for regression errors}
#'
#' \item{sigma}{Variance of the error}
#'
#' }
#' @export
LPDGP <- function(n, M, par) {
    # uniform[-1, 1] or beta
    X <- 2*(if (par$unif) stats::runif(n)
            else stats::rbeta(n=n, shape1=2, shape2=5)) -1

    ## Soft approximation to  (x)^2_{+}
    s0 <- function(x)
        if (par$k<Inf) -2*dilog(-exp(par$k*x))/par$k^2 else pmax(x, 0)^2

    ## symmetric and asymmetric spline
    sf <- function(X)
        X^2 + 2*(-s0(X-par$b1) - s0(-X-par$b1) + s0(X-par$b2) +
                 s0(-X-par$b2) - s0(X-par$b3) - s0(-X-par$b3))
    sfa <- function(X)
        (X+1)^2 +2*(-s0(X-par$b1) + s0(X-par$b2) - s0(X-par$b3) +
                    s0(X-par$b4))

    f <- if (!is.na(par$b4)) sfa(X)-sfa(0) else sf(X)-sf(0)
    ep <- par$sigma *
        (if (par$eps==1) stats::rnorm(n)
         else if (par$eps==2) (1+sqrt(abs(X)))*stats::rnorm(n)
         else (exp(stats::rnorm(n))-exp(1/2))/sqrt(exp(1)*(exp(1)-1)))
    data.frame(y=M/2 * f + ep, x=X)
}

## Compute confidence intervals and estimates using a variety of methods:
LPGetStats <- function(d, Mh) {
    ## CCF at CCF defaults
    if (!requireNamespace("nprobust", quietly = TRUE)) {
        stop("Please install nprobust package; it's needed
              for performing Monte Carlo.",
             call. = FALSE)
    }
    ## Set both h and b optimally
    d <- RDHonest::LPPData(d, point=0)
    bw.rot <- RDHonest::ROTBW.fit(d, kern="triangular", order=1, boundary=FALSE)
    c.rot <- nprobust::lprobust(y=d$Y, x=d$X, eval=0, kernel="tri", h=bw.rot,
                                rho=1, interior=TRUE, vce="nn")$Estimate[1, ]
    c.mse <- nprobust::lprobust(y=d$Y, x=d$X, eval=0, kernel="tri", vce="nn",
                                rho=0,
                                interior=TRUE, bwselect="mse-dpi")$Estimate[1, ]
    c.ce <- nprobust::lprobust(y=d$Y, x=d$X, eval=0, kernel="tri", vce="nn",
                               rho=0,
                               interior=TRUE, bwselect="ce-dpi")$Estimate[1, ]
    ## Set h=b (default in 0.1.4)
    c.mser <- nprobust::lprobust(y=d$Y, x=d$X, eval=0, kernel="tri", rho=1,
                                 vce="nn", h=c.mse["h"])$Estimate[1, ]
    c.cer <- nprobust::lprobust(y=d$Y, x=d$X, eval=0, kernel="tri", rho=1,
                                 vce="nn", h=c.ce["h"])$Estimate[1, ]

    r <- data.frame(method=rep(c("Conv", "RBC"), 5),
                    bw=rep(c("pMSE, rho=0", "pMSE, rho=1", "CE, rho=0",
                             "CE, rho=1", "ROT"), each=2),
                    bias=c(c.mse[c("tau.us", "tau.bc")],
                           c.mser[c("tau.us", "tau.bc")],
                           c.ce[c("tau.us", "tau.bc")],
                           c.cer[c("tau.us", "tau.bc")],
                           c.rot[c("tau.us", "tau.bc")]),
                    se=c(c.mse[c("se.us", "se.rb")],
                         c.mser[c("se.us", "se.rb")],
                         c.ce[c("se.us", "se.rb")],
                         c.cer[c("se.us", "se.rb")],
                         c.rot[c("se.us", "se.rb")]),
                    h=rep(c(c.mse["h"], c.mser["h"], c.ce["h"],
                            c.cer["h"], c.rot["h"]), each=2))
    ## pMSE and CE, rho=1 for conventional estimator are identical to rho=1
    r <- r[-c(3, 7), ]

    za <- qnorm(0.975)
    r$hl <- r$se*za
    r$r <- abs(r$bias/r$se) > za

    ## Honest inference:
    d1 <- RDHonest::NPRPrelimVar.fit(d, se.initial="EHW")
    if (!is.null(Mh)) {
        Ms <- c(Mh, RDHonest::NPR_MROT.fit(d))

        for (j in seq_along(Ms)) {
            h <- RDHonest::NPRHonest.fit(d1, M=Ms[j], sclass="H", J=3,
                                         opt.criterion="MSE", se.method="nn")
            Mlab <- if (j > length(Mh)) "ROT" else Ms[j]
            r <- rbind(r, data.frame(method=paste0("Hon, M=", Mlab),
                                     bw="MSE", bias=h$estimate,
                                     se=h$sd, h=h$hp, hl=h$hl,
                                     r=abs(h$estimate)>h$hl))
            ## RBC at this M
            rbc <- nprobust::lprobust(y=d$Y, x=d$X, eval=0, kernel="tri",
                                      interior=TRUE, h=h$hp, b=h$hp, vce="nn")
            sr <- rbc$Estimate[1, "se.rb"]
            br <- rbc$Estimate[1, "tau.bc"]
            r <- rbind(r, data.frame(method=paste0("RBC, M=", Mlab),
                                     bw="MSE, rho=1", bias=br, se=sr, h=h$hp,
                                     hl=sr*za, r=abs(br/sr) > za))
        }
    }
    rownames(r) <- NULL
    r
}

## Monte Carlo simulation of single design for coverage and length of CIs.
LPSimulateSingle <- function(n, M, par, nsim, cores=NULL, Mh) {
    ## get dims of output
    se <- LPGetStats(LPDGP(n, M, par), Mh)
    r <- array(dim=c(nrow(se), ncol(se)-2, nsim))
    set.seed(42)
    d <- LPDGP(n*nsim, M, par)
    Y <- matrix(d$y, nrow=n, ncol=nsim)
    X <- matrix(d$x, nrow=n, ncol=nsim)

    if (is.null(cores))
        cores <- parallel::detectCores()-1L     # number of cores

    SimDraw <- function(j)
        as.matrix(LPGetStats(data.frame(y=Y[, j], x=X[, j]),
                             Mh)[, -c(1, 2)])

    r <- if (cores>1) {
             parallel::mclapply(seq_len(nsim), SimDraw, mc.cores=cores)
         } else {
             lapply(seq_len(nsim), SimDraw)
         }
    r <- array(unlist(r), dim=c(nrow(r[[1]]), ncol(r[[1]]), length(r)))

    ra <- cbind(se[, 1:2], rowMeans(r, dims=2))

    names(ra) <- names(se)
    ## Add bandwidth stats
    bw <- r[, 3, ]
    stat <- function(f) sapply(seq_len(nrow(bw)), function(j) f(r[j, 3, ]))

    cbind(ra, data.frame(sdbw=stat(stats::sd), iqrbw=stat(stats::IQR),
                         medbw=stat(stats::median)))
}


#' Monte Carlo simulation for coverage and length of CIs for a set of designs
#'
#' Simulate CI coverage and length, comparing pointwise and uniform methods
#' using a number of designs
#' @param param list with following components:
#' \describe{
#'
#' \item{b1, b2, b3, b4}{Knots for spline. For symmetric designs, \code{p4=NA},
#' and the other parameters range in \eqn{[0, 1]}. For asymmetric designs, the
#' parameters range in \eqn{[-1, 1]}}
#'
#' \item{k}{Tuning parameter to approximate (x)^2_{+} by a smooth function}
#'
#' \item{unif}{\eqn{Unif(-1,1)} or scaled \eqn{Beta(2, 5)} distribution for running
#' variable?}
#'
#' \item{eps}{Integer specifying distribution for regression errors}
#'
#' \item{sigma}{Variance of the error}
#'
#' }
#' @param nsim number of simulation draws
#' @param eff.row Half-legnth comparisons should be wrt which row?
#' @param cores number of CPU cores (for parallel computation). Put
#'     \code{cores=0} for serial computation only.
#' @param Mh vector of values of M for Holder-honest CIs
#' @inheritParams LPDGP
#' @export
LPSimulate <- function(n=NULL, M, param, nsim, eff.row=1,
                       cores=NULL, Mh=NULL) {
    s <- Sys.time()
    if(is.null(cores))
        cores <- parallel::detectCores()-1L     # number of cores
    ## parallelize monte carlo or looping through designs?
    if(cores>1 & nrow(param)<20) {
        scores <- cores
        cores <- 1
    } else {
        scores <- 0
    }

    single <- function(j) {
        rj <- LPSimulateSingle(n=n, M=M, param[j, ], nsim=nsim, cores=scores, Mh)
        rj$eff <- rj$hl/rj$hl[eff.row]
        cbind(param[rep(j, nrow(rj)), ], rj)
    }

    r <- if (cores>1) {
             parallel::mclapply(seq_len(nrow(param)), single, mc.cores=cores)
         } else {
             lapply(seq_len(nrow(param)), single)
         }

    dt <- difftime(Sys.time(), s)
    print(dt)

    do.call(rbind, r)
}

#' Plot functions used in Monte Carlo design
#'
#' @param M bound on second derivative of f(X)
#' @param legends optional vector of legend labels
#' @inheritParams LPSimulate
#' @export
LPPlotDesign <- function(M, param, legends=NULL) {
    param$sigma <- 0
    d <- lapply(seq_len(nrow(param)), function(j) {
        r <- LPDGP(n=1000, M, par=param[j, ])
        r[order(r$x), ]
    })
    d <- do.call(rbind, d)
    param[, c("b1", "b2", "b3", "b4")] <-
        round(param[, c("b1", "b2", "b3", "b4")], 2)
    if (is.null(legends)) {
        legends <- sapply(seq_along(param$b1), function(j)
            paste0("$b_{1}=", param$b1[j], "$, $b_{2}=", param$b2[j], "$",
                   ", $b_{3}=", param$b3[j],
                   if (!is.na(param$b4[j]))
                       paste0("$, $b_{4}=", param$b4[j], "$") else "$"))
    }
    d <- data.frame(x=d$x, y=d$y,
                    what=rep(legends, each=nrow(d)/length(param$b1)))
    baseplot(d, ylab="$f(x)$", xlab="$x$",
             method=list(cex=0.9, "first.bumpup"))
}
