#' Compute \eqn{t^*}, optimal bias-variance trade-off.
#'
#' @param r Rate exponent
#' @param what what performance criterion to use. One of "FLCI" (fixed-length
#' CI), "OCI" (one-sided CI), "MSE" (mean-squared error)
#' @param alpha for "FLCI" and "OCI", determines confidence level,
#' \code{1-alpha}.
#' @param beta for "OCI", quantile of excess length to optimize
#' @keywords internal
tstar <- function(r, what="FLCI", alpha=0.05, beta=0.5)
    if (what=="FLCI") {
        stats::optimize(function(t) BWRisk(r, t, what="FLCI", alpha=alpha),
                        interval=c(0, 3), tol=.Machine$double.eps^0.75)$minimum
    } else if (what=="OCI") {
        (1/r-1)*(qnorm(beta)+qnorm(1-alpha))/2
    } else if (what=="MSE") {
        sqrt(1/r-1)
    }

#' Part of asymptotic risk that depends on bandwidth
#' @keywords internal
BWRisk <- function(r, t, what="FLCI", alpha=0.05, beta=0.5) {
    if (what=="FLCI") {
        t^(r-1) * RDHonest::CVb(t, alpha=alpha)$cv
    } else if (what=="OCI") {
        t^(r-1) * (2*t+qnorm(beta)+qnorm(1-alpha))
    } else if (what=="MSE") {
        t^(r-1) * sqrt(t^2+1)
    }
}



#' All plots in have this form
#'
#' @param data data frame with three columns: x, y, and what
#'     (color/linetype/facet specifier)
#' @param ylab,xlab Y- and X- axis labels
#' @param facet logical
#' @param br breaks for x axis
#' @param br.lab break labels for x axis
#' @param xlim X-axis limits
#' @param method method for \code{geom_dl} direct label
#' @keywords internal
baseplot <- function(data, ylab, xlab, br=NULL, br.lab=br,
                     xlim=NULL, rows=FALSE, cols=FALSE, method=NULL) {

    p <- if (rows & cols) {
             ggplot2::qplot(x=x, y=y, data=data, geom="line") +
                 ggplot2::facet_grid(rows=ggplot2::vars(rows),
                                     cols=ggplot2::vars(cols), scales="free")
         } else if (cols)
             ggplot2::qplot(x=x, y=y, data=data, geom="line") +
                 ggplot2::facet_grid(cols=ggplot2::vars(cols))
         else
             ggplot2::qplot(x=data$x, y=data$y, geom="line", color=data$what,
                            linetype=data$what)
    p <- p +
        ggplot2::xlab(xlab) + ggplot2::ylab(ylab) + ggplot2::theme_classic() +
        ggplot2::theme(legend.position = "none",
                       strip.background=ggplot2::element_rect(color = "white",
                           linetype=NULL),
                       axis.line.x=ggplot2::element_line(color="black",
                           size=0.2, linetype="solid"),
                       axis.line.y=ggplot2::element_line(color="black",
                           size=0.2, linetype="solid"))
    ## the axis.line.x and axis.line.y are needed as theme.classic has a bug in
    ## 2.1.0 version of ggplot2, see
    ## https://github.com/hadley/ggplot2/issues/1565

    if(!is.null(br))
        p <- p + ggplot2::scale_x_continuous(breaks=br, labels=br.lab)

    if(!is.null(xlim))
        p <- p + ggplot2::coord_cartesian(xlim = xlim)

    if (!is.null(method))
        p <-  p + directlabels::geom_dl(ggplot2::aes(label=what),
                                        data=data, method=method)
    p
}


#' Compute optimal bias-sd ratio \eqn{t^*} and MSE-bandwidth efficiency
#'
#' Computes optimal \eqn{t^*} for MSE, OCI, and FLCI, as well as relative
#' efficiency of confidence intervals centered around MSE-optimal bandwidth.
#'
#' @param r rate exponent (optimal rate of convergence). Can be a vector
#' @param beta For "OCI", quantiles of excess length to optimize. Can be a
#'     vector.
#' @param alpha Determines CI level for "FLCI" and "OCI", \code{1-alpha}. Can be
#'     a vector.
#' @return Returns data frame with following columns
#' \describe{
#'     \item{r}{Rate exponent as specified by \code{r}}
#'     \item{alpha}{Value of \eqn{\alpha} as specified by \code{alpha}}
#'     \item{beta}{Value of \eqn{\beta} as specified by \code{beta}}
#'     \item{criterion}{Efficiency criterion, MSE, OCI or FLCI}
#'     \item{t}{Optimal maximum bias-sd ratio, \eqn{t^*}}
#'     \item{MSEEfficiency}{Efficiency of minimax MSE-optimal bandwidth at this criterion}
#'     \item{TeXDescription}{LaTeX-friendly description of current row}
#' }
#' @examples
#' BWEfficiency(r=c(1/2, 2/3, 4/5), beta=0.5, alpha=c(0.1, 0.05))
#' @export
BWEfficiency <- function(r, beta, alpha) {

    ## Optimal t
    mse <- data.frame(r=r, alpha=NA, beta=NA, criterion="MSE",
                      t=sapply(r, function(r) tstar(r, what="MSE")),
                      MSEEfficiency=1)
    fl <- expand.grid(r=r, alpha=alpha, beta=NA, criterion="FLCI")
    fl$t <- apply(fl[, -4], 1, function(u)
        tstar(r=u["r"], what="FLCI", alpha=u["alpha"]))
    os <- expand.grid(r=r, alpha=alpha, beta=beta, criterion="OCI")
    os$t <- apply(os[, -4], 1, function(u)
        tstar(r=u["r"], what="OCI", alpha=u["alpha"], beta=u["beta"]))

    ## Efficiency of MSE
    eff <- function(u, what) {
        BWRisk(u["r"], u["t"], what=what, alpha=u["alpha"], beta=u["beta"]) /
            BWRisk(u["r"], mse[mse$r==u["r"], "t"],
                   what=what, alpha=u["alpha"], beta=u["beta"])
    }

    fl$MSEEfficiency <- apply(fl[, -4], 1, function(u) eff(u, what="FLCI"))
    os$MSEEfficiency <- apply(os[, -4], 1, function(u) eff(u, what="OCI"))

    mse$TeXDescription <- "MSE"
    fl$TeXDescription <- paste("FLCI, $\\alpha=", fl$alpha, "$", sep="")
    os$TeXDescription <- paste("OCI,  $\\alpha=", os$alpha, "$, $\\beta=",
                               os$beta, "$", sep="")
    rbind(mse, fl, os)
}

#' Figures of optimal bias-variance trade-off and MSE efficiency
#'
#' Makes plots of optimal FLCI and OCI bias-sd ratios as function of the rate
#' exponent (one plot each), and plots of efficiency of two-sided and one-sided
#' CIs based on MSE-optimal bandwdith relative to FLCI/OCI optimal bandwidth as
#' function of the rate exponent (one plot each)
#'
#' @return Four ggplot2 figures.
#' @export
plot_BWEfficiency <- function() {
    d <- BWEfficiency(seq(0.5, 0.999, by=0.01), c(0.5, 0.8), c(0.05, 0.01))
    d2 <- BWEfficiency(seq(0.5, 0.999, by=0.01), c(0.5, 0.8),
                       c(0.1, 0.05, 0.01))
    br <- c(seq(0.5, 1, by=0.1))        # breaks in plot

    ## Fixed-length CI plots
    dfl <- d2[!grepl("OCI", d2$criterion), c("r", "t", "TeXDescription")]
    dos <- d[!grepl("FL", d$criterion), c("r", "t", "TeXDescription")]
    dfl.e <- d2[d2$criterion=="FLCI", c("r", "MSEEfficiency", "TeXDescription")]
    dos.e <- d[d$criterion=="OCI", c("r", "MSEEfficiency", "TeXDescription")]

    names(dfl) <- names(dos.e) <- names(dfl.e) <- names(dos) <-
        c("x", "y", "what")

    m1 <- list(cex=0.9, "first.qp")
    p1 <- baseplot(dfl, "", "$r$", br, br, c(0.35, 1), method=m1)
    p2 <- baseplot(dos, "", "$r$", br, br, c(0.25, 1), method=m1)
    p3 <- baseplot(dfl.e, "Relative efficiency", "$r$", br, br, c(0.3, 1),
                   method=m1)
    p4 <- baseplot(dos.e, "Relative efficiency", "$r$", br, br, c(0.5, 1.2),
                   method=list(cex=0.9, "last.qp"))

    list(p1=p1, p2=p2, p3=p3, p4=p4)
}
