#' Find interval containing zero of a function, then find the zero
#'
#' Given function \code{f} find \code{x0} such that \code{f(x0)==0}
#' @param f function whose root we're looking for
#' @param ival upper endpoint of initial interval in which to search
#' @param negative logical: should the lower endpoint be \code{1/ival} (if the
#'     root is guaranteed to be positive), or \code{-ival}?
#' @keywords internal
FindZero <- function(f, ival=1.1, negative=TRUE) {
    minval <- function(ival) if (negative==TRUE) -ival else min(1/ival, 1e-3)

    while(sign(f(ival))==sign(f(minval(ival))))
            ival <- 2*ival
    stats::uniroot(f, c(minval(ival), ival), tol=tol)$root
}


#' Replicate empirical analysis
#'
#' @param Ms vector of smoothness parameters M, such that the second derivative
#'     is bounded by M (i.e. \eqn{M=2C} in finite-sample paper)
#' @param bws bandwidths
#' @inheritParams RDHonest::RDHonest.fit
#' @export
HolderAnalysis <- function(d, kern, Ms=NULL, bws=NULL, se.method="nn",
                           J=3, alpha=0.05) {
    fs <- data.frame()
    d <- RDHonest::NPRPrelimVar.fit(d, se.initial="EHW")

    ## Finite-sample optimal bw given M
    rdh <- function(M, ...)
        RDHonest::NPRHonest.fit(d, M, kern, sclass="H",
                                J=J, alpha=alpha, ...)

    if (!is.null(bws)) {
        for (b in bws) {
            ## Finite-sample method. For uniform kernel, objective is piecewise
            ## constant
            Est <- function(M) rdh(M, opt.criterion="MSE", bw.equal=TRUE,
                                   se.method="supplied.var")$hp-b
            M2 <- FindZero(Est, negative=FALSE)
            fs <- rbind(fs, as.data.frame(c(rdh(M2, h=b,
                                                se.method=se.method)[-c(2, 12)],
                                            list(M=M2))))
        }
    } else if (!is.null(Ms)) {
        for (M in Ms) {
            Est <- rdh(M, opt.criterion="MSE", bw.equal=TRUE,
                       se.method=se.method)[-c(2, 12)]
            fs <- rbind(fs, as.data.frame(c(Est, list(M=M))))
        }
    }
    fs$cvb <- RDHonest::CVb(fs$maxbias/fs$sd, alpha)$cv

    fs
}
